import timeit
import os
import os.path
import  csv
import json
from .MeasureManager import MeasureManager
import numpy as np
from termcolor import colored
class Measure(object):
    
    def calculate(self):
        return 'Finish caculating'
    def LoaddistMatrixAuto(self):
        if MeasureManager.IS_LOAD_AUTO == False or MeasureManager.CURRENT_DATASET=="None":
            print( f'SKIP LOADING distMatrix because IS_LOAD_AUTO={MeasureManager.IS_LOAD_AUTO} or dbname={MeasureManager.CURRENT_DATASET}; If you cluster a dataset multiple times, make sure to include the \'dbname\' parameter during initialization to cache the distance matrix of the dataset. Calculating distance matrix can take a lot of time with high categirical values dataset.' )
            return False
        path = 'saved_dist_matrices/json/' + self.name+"_" + MeasureManager.CURRENT_DATASET+ ".json"
        if os.path.isfile(path):
            with open(path, "r") as fp:
                self.distMatrix = json.load(fp)
        else: 
            print(colored('CANNOT OPEN FILE: ' +path),'yellow' )
            return False

        print("Loaded dist matrix: " , path)
        return True
    def SavedistMatrix(self):
        if not os.path.exists('saved_dist_matrices'):
            os.makedirs('saved_dist_matrices')
        if not os.path.exists('saved_dist_matrices/json'):
            os.makedirs('saved_dist_matrices/json')
        path = 'saved_dist_matrices/json/' + self.name+"_" + MeasureManager.CURRENT_DATASET+ ".json"
        with open('saved_dist_matrices/' + self.name+"_" + MeasureManager.CURRENT_DATASET, 'w') as f:
            wr = csv.writer(f)
            wr.writerows(self.distMatrix)
        with open(path, "w") as fp:
            json.dump(self.distMatrix, fp)
        print("Saving",self.name,"to:" ,path)
    def setUp(self, X, y):
        self.X_ = X
        self.y_ = y
        return 0
    def test(self):
        print('Test OK')
    def GeneratedistMatrix(self):
        D = len(self.X_[0])
        self.max = []
        for i in range(len(self.X_[0])):
            self.max.append(max(self.X_[:,i]))
        self.distMatrix = [];
        for d in range(D):
            matrix2D = [] # 2D array for 1 dimension
            for i in range(self.max[d]+1):
                matrix1D = [] # 1D array for 1 dimension
                for j in range(self.max[d]+1): 
                    matrix_tmp = self.CalcdistanceArrayForDimension(d,i,j)
                    matrix1D.append(matrix_tmp)
                matrix2D.append(matrix1D)
            self.distMatrix.append(matrix2D)

    def GeneratesimMatrix(self):
        self.d = d = len(self.X_[0])
        self.simMatrix = [];
        self.D = D = [len(np.unique(self.X_[:,i])) for i in range(d) ]
        self.medianValues = {}
        for di in range(d):
            
            matrix2D = [] # 2D array for 1 dimension
            for i in range(D[di]):
                matrix1D = [] # 1D array for 1 dimension
                for j in range(D[di]):
                    #matrix_tmp = 1-self.distMatrix[di][i][j]
                    if self.distMatrix[di][i][j] ==0: matrix_tmp = 10000;
                    else : matrix_tmp= 1/self.distMatrix[di][i][j]

                    matrix1D.append(matrix_tmp)
                matrix2D.append(matrix1D)
            self.simMatrix.append(matrix2D)



#from Measures.Overlap import Overlap

class MeasureManager(object):
    CURRENT_MEASURE = "None"
    CURRENT_DATASET = "None"
    IS_LOAD_AUTO = False
    MEASURE_LIST = ['OF','Overlap','New_Lin','DILCA','Eskin','IOF','Nguyen_Huynh','Goodall1']
    DATASET_LIST = ['soybean_small.csv','balance-scale.csv','audiology_c.csv','zoo_c.csv','tae_c.csv','post-operative.csv','hayes-roth_c.csv','dermatology_c.csv','soybean.csv','nursery.csv','connect.csv','chess.csv','breast.csv','car.csv','mushroom.csv','splice.csv','tictactoe.csv','vote.csv','flare.csv','lymph.csv','lung.csv']
    DATASET_LIST_MK = ['soybean_small.csv','balance-scale.csv','balance-scale.csv','zoo_c.csv','tae_c.csv','post-operative.csv','hayes-roth_c.csv','dermatology_c.csv','soybean.csv','nursery.csv','connect.csv','chess.csv','breast.csv','car.csv','mushroom.csv','splice.csv','tictactoe.csv','vote.csv','flare.csv','lymph.csv','lung.csv']
    DATASET_LIST_BIG = ['big.csv']

    def Init(self):
        print("OK")




#Note by toanstt: The short lists of near clusters are updated each iterator

import os
import os.path
import sys
from sys import platform
sys.path.append(os.path.join(os.getcwd(), "Measures"))
sys.path.append(os.path.join(os.getcwd(), "LSH"))
sys.path.append(os.path.join(os.getcwd(), "../"))
sys.path.append(os.path.join(os.getcwd(), "../Dataset"))
sys.path.append(os.path.join(os.getcwd(), "../Measures"))
sys.path.append(os.path.join(os.getcwd(), "../LSH"))
sys.path.append(os.path.join(os.getcwd(), "./ClusteringAlgorithms"))
import numpy as np
import pandas as pd
#from kmodes_lib import KModes

from collections import defaultdict
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_array
import timeit
#from kmodes.util import get_max_value_key, encode_features, get_unique_rows, \
#    decode_centroids, pandas_to_numpy

from .ClusteringAlgorithm import ClusteringAlgorithm
from .LSHkRepresentatives import CategoricalDatasetNormalization
from sklearn.metrics.cluster import adjusted_rand_score
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.metrics.cluster import adjusted_mutual_info_score
from sklearn.metrics.cluster import homogeneity_score
import random
from .LSH import LSH 

import multiprocessing as mp
import math
from collections import defaultdict


class LSHkRepresentatives_Full(ClusteringAlgorithm):

    def SetupLSH(self, hbits=-1,k=-1,measure='DILCA' ):
        self.d = self.d_CATE
        self.D = self.D_CATE
        self.X = self.X_CATE
        self.uniques = self.uniques_CATE
        if self.d_NUM > 0:
            print("\n\n YOU ARE SETTING SOME NUMERIC ATTRIBUTES IN YOUR DATA, SO PLEASE USE LSHkPrototypes INSTEAD OF LSHkRepresentatives_Full \n\n")


        self.CheckNormalizedData()
        #hbits = 2*math.ceil(math.log(self.k)/math.log(2))
        start = timeit.default_timer()
        self.lsh = LSH(self.X,self.y,measure=measure,hbits=hbits)
        self.lsh.DoHash()
        #self.lsh.CorrectSingletonBucket()
        self.time_lsh = timeit.default_timer() - start
        self.AddVariableToPrint("Time_lsh",self.time_lsh)
        return self.time_lsh
    def SetupMeasure(self, classname):
        self.measurename = classname
        module = __import__(classname, globals(), locals(), ['object'])
        class_ = getattr(module, classname)
        self.measure = class_()
        self.measure.setUp(self.X, self.y)
    def test(self):
        print("a234 " + str(self.k))
    def Distance(self,representative, point):
        sum=0;
        for i in range (self.d):
            sum = sum + representative[i][point[i]]
        return self.d - sum
    
    def MovePoint(self, point_id, from_id, to_id ,representatives_count, representatives_sum,membship, curpoint,labels_matrix):
        labels_matrix[point_id] = to_id
        membship[to_id, point_id] = 1
        membship[from_id, point_id] = 0
        representatives_sum[to_id]+=1
        representatives_sum[from_id]-=1 
        for ii, val in enumerate(curpoint):
            representatives_count[to_id][ii][val]+=1
            representatives_count[from_id][ii][val]-=1
    def CheckEmptyClusters(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        move =0
        big_cluster_id = -1
        for ki in range(self.k):
            if representatives_sum[ki] ==0 :
                time_start_found = timeit.default_timer()
                big_cluster_id = np.argmax([sum(mem_) for mem_ in membship])
                choices = [i for i in range(self.n) if membship[big_cluster_id][i] == 1 ]
                rindx = int(self.farest_point[big_cluster_id][0])
                self.MovePoint(rindx, big_cluster_id,ki, representatives_count, representatives_sum,membship,self.X[rindx],labels_matrix  )
                move +=1
                if TDef.verbose >=2 :  print("Found a empty cluster (Loop): ", ki, " Timelapse=", "%.2f"%(timeit.default_timer()-time_start_found))
        return move
    def CheckEmptyClusters_Init(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        count =0;
        move =0
        big_cluster_id = -1
        is_have = False
        for ki in range(self.k):
            if representatives_sum[ki] ==0 :
                count+=1
                is_have = True
                
                rindx = random.randint(0, self.n-1) 
                big_cluster_id = labels_matrix[rindx]
                self.MovePoint(rindx, big_cluster_id,ki, representatives_count, representatives_sum,membship,self.X[rindx],labels_matrix  )
                move +=1
        #if TDef.verbose >=3:  print("WARING: Found " + str(count) +"/" + str(self.k) + " empty clusters (Init) ")
        if is_have: return move + self.CheckEmptyClusters_Init(representatives, X,representatives_sum, representatives_count,membship,labels_matrix)
        return move
    def InitClusters(self,representatives,representatives_sum,representatives_count):
        for ki in range(self.k):
            for i in range(self.d):
                sum_ = 0
                for j in range(self.D[i]): sum_ = sum_ + representatives[ki][i][j]
                for j in range(self.D[i]): representatives[ki][i][j] = representatives[ki][i][j]/sum_;

    def DistanceRepresentativestoAPoints(self,representatives, point):
        dist_matrix = [self.Distance(c, point) for c in representatives]
        representative_id = np.argmin(dist_matrix)
        return representative_id, dist_matrix[representative_id]

    
    def UpdateLabelsInit(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        cost = 0
        move = 0
        self.preferList= defaultdict(set)
        for ipoint, curpoint in enumerate(X):
            representative_id,tmp = self.DistanceRepresentativestoAPoints(representatives, curpoint)
            cost += tmp
            labels_matrix[ipoint] = representative_id
            membship[representative_id, ipoint] = 1
            representatives_sum[representative_id]+=1
            for ii, val in enumerate(curpoint):
                representatives_count[representative_id][ii][val]+=1
            self.preferList[self.lsh.hash_values[ipoint]].add(labels_matrix[ipoint])
        self.CheckEmptyClusters(representatives, X,representatives_sum, representatives_count,membship,labels_matrix)
        self.dist_matrix_tmp = [1000000000 for i in range(self.k)]
        return cost ,move, 0
    def UpdateLabelsLast(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        cost = 0
        self.preferList= defaultdict(set)


        for ipoint, curpoint in enumerate(X):
            representative_id,tmp = self.DistanceRepresentativestoAPoints(representatives, curpoint)
            cost += tmp
            membship[representative_id, ipoint] = 1
        return cost ,0, 0

    def DistanceRepresentativestoAPoints_LSH(self,item_id, point,labels_matrix,representatives):
        myset = self.preferList[self.lsh.hash_values[item_id]]
        dist_min = 1000000000
        dist_index =-1
        for i in myset:
            dist = self.Distance(representatives[i], point)
            if dist_min > dist:
                dist_min = dist 
                dist_index = i
        return dist_index, dist_min 
    def DistanceRepresentativestoAPoints_LSH2(self,item_id, point,labels_matrix,representatives):
        #myset = self.preferList[self.lsh.hash_values[item_id]]
        #myset = self.near_clusters[self.lsh_group[item_id]]
        myset = self.near_clusters[labels_matrix[item_id]]

        dist_min = 1000000000
        dist_index =-1
        for i in myset:
            dist = self.Distance(representatives[i], point)
            if dist_min > dist:
                dist_min = dist 
                dist_index = i
        return dist_index, dist_min 

    def UpdateLabels(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        cost = 0
        move = 0
        for i in range(self.k):
            self.farest_point[i][1] = 0

        for ipoint, curpoint in enumerate(X):
            #representative_id,tmp = self.DistanceRepresentativestoAPoints_LSH(ipoint, curpoint,labels_matrix,representatives)
            #representative_id,tmp = self.DistanceRepresentativestoAPoints(representatives, curpoint)
            representative_id,tmp = self.DistanceRepresentativestoAPoints_LSH2(ipoint, curpoint,labels_matrix,representatives)
            if tmp > self.farest_point[representative_id][1]:
                self.farest_point[representative_id][1] = tmp
                self.farest_point[representative_id][0] = ipoint
            cost += tmp
            
            if membship[representative_id, ipoint]: continue
            old_clust = labels_matrix[ipoint]
            self.MovePoint(ipoint, old_clust,representative_id, representatives_count, representatives_sum,membship,curpoint,labels_matrix  )
            move +=1
        #Check empty clusters
        move  += self.CheckEmptyClusters(representatives, X,representatives_sum, representatives_count,membship,labels_matrix)
        return cost ,move, 0

    def UpdateRepresentatives(self,representatives,representatives_sum,representatives_count ) :  
        for ki in range(self.k):
            for di in range(self.d):
                    for vj in range(self.D[di]):
                        representatives[ki][di][vj] =  representatives_count[ki][di][vj]/representatives_sum[ki]
        return 0
    def GetLabels(self, membship):
        labels = np.empty(self.n, dtype=np.uint16)
        for ki in range(self.k):
            for i in range(self.n):
                if membship[ki][i]:
                    labels[i] = ki
        return labels
    def UpdateClusterDistMatrix(self,representatives):
        for i in range(self.k):
            for j in range(i+1,self.k):
                sum =0
                for k in range(self.d):
                    sum+= np.linalg.norm(representatives[i][k] - representatives[j][k])
                self.cluster_dist_matrix[i][j] = self.cluster_dist_matrix[j][i] =  sum
            self.near_clusters[i] = np.argsort(self.cluster_dist_matrix[i])[0:self.nearcluster_count+1]
        #print(self.near_clusters[0])
    def CheckNormalizedData(self):
        isNormalized = True
        for d in range(self.d):
            if isNormalized == False: break
            for dvalue in range(self.D[d]):
                if dvalue not in self.uniques[d]:
                    isNormalized = False; break
        #print("\n\n\n ",isNormalized,"\n\n\n\n\n")
        if isNormalized== False:
            self.normalizer = CategoricalDatasetNormalization(self.X)
            self.X = self.normalizer.Normalize(self.X)
        else: self.normalizer = None
    def DoCluster(self,n_group=2,init_type ='random'):
        n = len(self.X)
        d = self.X.shape[1]
        if n_group==-1:
            if self.k >= 8: n_group = self.k/4;
            else: n_group=2
        self.AddVariableToPrint("init_type",init_type)
        self.AddVariableToPrint("n_group",n_group)
        self.k = k = n_clusters = self.k
        self.farest_point = np.zeros((self.k,2))

        self.name = "LSHkRepresentatives_Full"
        
        
        # self.n = n = self.X.shape[0];
        # self.d = d = X.shape[1]
        # self.D = D = [len(np.unique(X[:,i])) for i in range(d) ]
        all_labels = []
        all_costs = []
        start_time = timeit.default_timer()
        
        self.cluster_dist_matrix = np.zeros((self.k,self.k))
        self.near_clusters = [[] for j in range(self.k)]

        print(" n_group=" + str(n_group) + " Average neighbors:" + str(self.k/n_group) )
        for init_no in range(self.n_init):
            start_time_init = timeit.default_timer()
            #if TDef.verbose >=1: print (self.name +' Init ' + str(init_no))
            self.random_state = check_random_state(None)
            membship = np.zeros((k, n), dtype=np.uint8)
            labels_matrix = np.empty(n, dtype=np.uint16)
            
            #if TDef.verbose >=3: print("Initing memories");
            for i in range(n): labels_matrix[i] = 65535
            representatives_count = [[[0 for i in range(self.D[j])] for j in range(d)]for ki in range(k)]
            representatives_sum = [0 for ki in range(k)]
            last_cost = float('inf')

            #representatives = [[[0 for i in range(D[j])] for j in range(d)] for ki in range(k)]
            representatives = [[np.zeros(self.D[j]) for j in range(d)] for ki in range(k)]

            buckets = [(k,len(self.lsh.hashTable[k])) for k in self.lsh.hashTable.keys()]
            buckets2 = sorted(buckets, key=lambda x: -x[-1])

            masterkeys=[]
            n_group = int(self.k/n_group)
            self.nearcluster_count=n_group
            for i in range( self.k - len(buckets2)):
                buckets2.append((0,0))
            for i in range(self.k):
                masterkeys.append(buckets2[i][0])
            #if TDef.verbose >=3: print("Computing dist matrices");
            dist_from_master_to_master = np.zeros((self.k,self.k))

            for i in range(self.k):
                for j in range(i+1,self.k):
                    dist_from_master_to_master[i][j]= dist_from_master_to_master[j][i] = \
                    self.lsh.hammingDistance(masterkeys[i],masterkeys[j])

            k2 = len(buckets2)
            for i in range(k2):
                buckets = self.lsh.hashTable[buckets2[i][0]]
                if i < k: i2= i
                else : 
                    if init_type== 'best':
                        i2 = -1; dst= float('inf')
                        for i3 in range(k):
                            dst_tmp = self.lsh.hammingDistance(buckets2[i][0],masterkeys[i3])
                            if dst_tmp<=dst: dst = dst_tmp; i2=i3
                    else: i2 = random.randint(0,self.k-1)
                for j in buckets:
                    representatives_sum[i2]+=1
                    labels_matrix[j] = i2
                    membship[i2][j] = 1
                    for ii, val in enumerate(self.X[j]):
                        representatives_count[i2][ii][val]+=1

            #if TDef.verbose >=3: print ('Checking Empty Clusters')
            self.CheckEmptyClusters_Init(representatives, self.X,representatives_sum, representatives_count,membship,labels_matrix)
            #if TDef.verbose >=3: print ('UpdateRepresentatives')
            self.UpdateRepresentatives(representatives,representatives_sum,representatives_count ) ;
            self.UpdateClusterDistMatrix(representatives)
            #if TDef.verbose >=1: print (' Init timelapse: ' + str(timeit.default_timer() -start_time_init))
            for i in range(self.n_iter):
                start_time_iter = timeit.default_timer()
                self.iter = i
                cost , move, count_empty = self.UpdateLabels(representatives, self.X,representatives_sum, representatives_count,membship,labels_matrix)
                self.UpdateRepresentatives(representatives,representatives_sum,representatives_count ) ;
                self.UpdateClusterDistMatrix(representatives)
                #if TDef.verbose >=2: print ('Iter ' + str(i),"Cost: ", "%.2f"%cost," Move:", move, " Num empty:", count_empty," Timelapse:","%.2f"% (timeit.default_timer()-start_time_iter ))
                if last_cost == cost and move==0: 
                    last_cost,_,_ = self.UpdateLabelsLast(representatives, self.X,representatives_sum, representatives_count,membship,labels_matrix)
                    #if TDef.verbose >=2: print ('Iter (last)' ,"Cost: ", "%.2f"%last_cost," Move:", move, " Num empty:", count_empty," Timelapse:","%.2f"% (timeit.default_timer()-start_time_iter ))
                    break 
                last_cost = cost


            labels = self.GetLabels(membship)
            all_costs.append(cost)
            all_labels.append(labels)
            
        best = np.argmin(all_costs)
        labels = all_labels[best]
        self.time_score = (timeit.default_timer() - start_time)/ self.n_init
        self.labels = labels
        print("LSH time:", self.time_lsh ,"Score: ", all_costs[best] , " Time:", self.time_score)
        self.scorebest = all_costs[best]
        self.representatives = representatives
        return self.labels
        # Update representives
    def predict(self,x):
        if self.normalizer!=None:
            x = self.normalizer.Normalize(x)
        if len(x.shape) ==1:
            dist_matrix = self.DistanceRepresentativestoAPoints(self.representatives, x)
            return dist_matrix[0]
        else:
            out = []
            for xi in x:
                out.append(self.DistanceRepresentativestoAPoints(self.representatives, xi)[0])
            return np.array(out)
    
def Test(): 
    #MeasureManager.CURRENT_DATASET = 'tae_c.csv'
    MeasureManager.CURRENT_DATASET = 'lung.csv'
    MeasureManager.CURRENT_MEASURE = 'Overlap'
    if TDef.data!='': MeasureManager.CURRENT_DATASET = TDef.data
    if TDef.measure!='': MeasureManager.CURRENT_MEASURE = TDef.measure
    if TDef.test_type == 'syn':
        DB = tulti.LoadSynthesisData(TDef.n,  TDef.d, TDef.k)
        MeasureManager.CURRENT_DATASET= DB['name']
    else:
        DB = tulti.LoadRealData(MeasureManager.CURRENT_DATASET)
    print("\n\n############## LSHkRepresentatives_Full ###################")
    lshkrepresentatives = LSHkRepresentatives_Full(DB['DB'],DB['labels_'] ,dbname=MeasureManager.CURRENT_DATASET ,k=TDef.k)
    lshkrepresentatives.SetupMeasure(MeasureManager.CURRENT_MEASURE)
    lshkrepresentatives.SetupLSH(measure=MeasureManager.CURRENT_MEASURE)
    lshkrepresentatives.DoCluster(n_group=TDef.n_group,init_type=TDef.init_type)
    lshkrepresentatives.CalcScore()

def TestDatasets(): 
    for dbname in MeasureManager.DATASET_LIST:
        DB = tulti.LoadRealData(dbname)
        MeasureManager.CURRENT_DATASET = dbname
        MeasureManager.CURRENT_MEASURE = 'Overlap'
        print("\n\n############## "+self.name+" ###################")
        alo = LSHkRepresentatives_Full(DB['DB'],DB['labels_'],dbname=MeasureManager.CURRENT_DATASET )
        alo.SetupMeasure(MeasureManager.CURRENT_MEASURE)
        alo.SetupLSH(measure=MeasureManager.CURRENT_MEASURE)
        alo.DoCluster()
        alo.CalcScore()

def TestMeasures():
    table = MyTable()
    for measure in MeasureManager.MEASURE_LIST:
        MeasureManager.CURRENT_MEASURE = measure
        for dataset in MeasureManager.DATASET_LIST:
            MeasureManager.CURRENT_DATASET=dataset
            DB = tulti.LoadRealData(MeasureManager.CURRENT_DATASET)
            alo = LSHkRepresentatives_Full(DB['DB'],DB['labels_'],dbname=MeasureManager.CURRENT_DATASET )
            alo.SetupMeasure(MeasureManager.CURRENT_MEASURE)
            alo.SetupLSH(measure=MeasureManager.CURRENT_MEASURE)
            alo.DoCluster()
            alo.CalcScore()

            alo.AddValuesToMyTable(table,measure == MeasureManager.MEASURE_LIST[0],dataset)
            table.SaveToExcelFolder("RESULT/r20200916" + alo.name,TDef.o,[i+1 for i in range(0,1000)])

if __name__ == "__main__":
    TDef.InitParameters(sys.argv)
    if TDef.test_type == 'datasets':
        TestDatasets()
    elif TDef.test_type == 'measures':
        TestMeasures()
    else:
        Test()


import os
import os.path
import sys
from sys import platform

import numpy as np
import pandas as pd
from collections import defaultdict
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_array
import timeit
# from kmodes.util import get_max_value_key, encode_features, get_unique_rows, \
#     decode_centroids, pandas_to_numpy

from .ClusteringAlgorithm import ClusteringAlgorithm
from sklearn.metrics.cluster import adjusted_rand_score
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.metrics.cluster import adjusted_mutual_info_score
from sklearn.metrics.cluster import homogeneity_score
import random
from .LSH import LSH 
import multiprocessing as mp
import math
from collections import defaultdict
from .LSHkRepresentatives import CategoricalDatasetNormalization

        
class LSHkPrototypes(ClusteringAlgorithm):

    def SetupLSH(self, hbits=-1,k=-1,measure='DILCA' ):
        self.CheckNormalizedData()
        start = timeit.default_timer()
        self.lsh = LSH(self.X_CATE,self.y,measure=measure,hbits=hbits)
        self.lsh.DoHash()
        self.time_lsh = timeit.default_timer() - start
        self.AddVariableToPrint("Time_lsh",self.time_lsh)
        return self.time_lsh
    def SetupMeasure(self, classname):
        self.measurename = classname
        module = __import__(classname, globals(), locals(), ['object'])
        class_ = getattr(module, classname)
        self.measure = class_()
        self.measure.setUp(self.X, self.y)
    def test(self):
        print("a234 " + str(self.k))

    def Distance_CATE(self,representative, point):
        sum=0
        for i in range (self.d_CATE):
            sum = sum + representative[i][point[i]]
        return self.d_CATE - sum
    def Distance_NUM(self,center, point):
        return np.linalg.norm(center-point)

    def Distance_2(self,representative,mean, pointCATE, pointNUM):
        sum=0
        for i in range (self.d_CATE):
            sum = sum + representative[i][pointCATE[i]]
        sum =  self.d_CATE - sum

        return np.linalg.norm(pointCATE-pointNUM)*self.weight_NUM + sum*self.weight_CATE

    
    def MovePoint(self, point_id, from_id, to_id ,representatives_count, representatives_sum,membship, curpoint,labels_matrix,means_count,means_sum):
        labels_matrix[point_id] = to_id
        membship[to_id, point_id] = 1
        membship[from_id, point_id] = 0

        representatives_sum[to_id]+=1
        representatives_sum[from_id]-=1 

        for ii, val in enumerate(curpoint):
            representatives_count[to_id][ii][val]+=1
            representatives_count[from_id][ii][val]-=1


        means_count[to_id] += 1
        means_count[from_id] -= 1

        if  len(means_sum.shape)>1:
            for di in range(self.d_NUM):
                means_sum[to_id][di] += self.X_NUM[point_id][di]
                means_sum[from_id][di] -= self.X_NUM[point_id][di]

    def CheckEmptyClusters(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_sum,means):
        move =0
        big_cluster_id = -1
        for ki in range(self.k):
            if representatives_sum[ki] ==0 :
                big_cluster_id = np.argmax([sum(mem_) for mem_ in membship])
                choices = [i for i in range(self.n) if membship[big_cluster_id][i] == 1 ]
                rindx = int(self.farest_point[big_cluster_id][0])
                self.MovePoint(rindx, big_cluster_id,ki, representatives_count, representatives_sum,membship,self.X_CATE[rindx],labels_matrix,means_count,means_sum  )
                move +=1

        return move
    def InitClusters(self,representatives,representatives_sum,representatives_count):
        for ki in range(self.k):
            for i in range(self.d):
                sum_ = 0
                for j in range(self.D[i]): sum_ = sum_ + representatives[ki][i][j]
                for j in range(self.D[i]): representatives[ki][i][j] = representatives[ki][i][j]/sum_;

    def DistancePrototypestoAPoints(self,representatives,means, point_CATE,point_NUM):
        dist_matrix_CATE = [self.Distance_CATE(c, point_CATE) for c in representatives]
        dist_matrix_NUM = [self.Distance_NUM(c, point_NUM) for c in means]
        dist_matrix = np.add(np.multiply(dist_matrix_CATE,self.weight_CATE),np.multiply(dist_matrix_NUM,self.weight_NUM))
        representative_id = np.argmin(dist_matrix)
        return representative_id, dist_matrix[representative_id]

    def UpdateLabelsInit(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_sum,means):
        cost = 0
        move = 0
        self.preferList= defaultdict(set)
        for ipoint, curpoint in enumerate(X):
            representative_id,tmp = self.DistancePrototypestoAPoints(representatives,means, curpoint, self.X_NUM[ipoint])
            cost += tmp
            labels_matrix[ipoint] = representative_id
            membship[representative_id, ipoint] = 1
            representatives_sum[representative_id]+=1
            for ii, val in enumerate(curpoint):
                representatives_count[representative_id][ii][val]+=1
            self.preferList[self.lsh.hash_values[ipoint]].add(labels_matrix[ipoint])
        self.CheckEmptyClusters(representatives, X,representatives_sum, representatives_count,membship,labels_matrix)
        self.dist_matrix_tmp = [1000000000 for i in range(self.k)]
        return cost ,move, 0
    def UpdateLabelsLast(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_sum,means):
        cost = 0
        self.preferList= defaultdict(set)


        for ipoint, curpoint in enumerate(X):
            representative_id,tmp = self.DistancePrototypestoAPoints(representatives,means, curpoint, self.X_NUM[ipoint])
            cost += tmp
            membship[representative_id, ipoint] = 1
        return cost ,0, 0

    

    def UpdateLabels(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_sum,means):
        cost = 0
        move = 0
        for i in range(self.k):
            self.farest_point[i][1] = 0

        for ipoint, curpoint in enumerate(X):
            representative_id,tmp = self.DistancePrototypestoAPoints(representatives,means, curpoint, self.X_NUM[ipoint])
            if tmp > self.farest_point[representative_id][1]:
                self.farest_point[representative_id][1] = tmp
                self.farest_point[representative_id][0] = ipoint
            cost += tmp
            
            if membship[representative_id, ipoint]: continue
            old_clust = labels_matrix[ipoint]
            self.MovePoint(ipoint, old_clust,representative_id, representatives_count, representatives_sum,membship,curpoint,labels_matrix,means_count,means_sum  )
            move +=1
        #Check empty clusters
        move  += self.CheckEmptyClusters(representatives, X,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_sum,means)
        return cost ,move, 0

    def UpdateRepresentatives(self,representatives,representatives_sum,representatives_count,means_count,means_sum,means ) :  
        for ki in range(self.k):
            for di in range(self.d_CATE):
                for vj in range(self.D_CATE[di]):
                    representatives[ki][di][vj] =  representatives_count[ki][di][vj]/representatives_sum[ki]
            
            for di in range(self.d_NUM):
                means[ki][di] = means_sum[ki][di]/means_count[ki]
                
        return 0
    def GetLabels(self, membship):
        labels = np.empty(self.n, dtype=np.uint16)
        for ki in range(self.k):
            for i in range(self.n):
                if membship[ki][i]:
                    labels[i] = ki
        return labels
    def CheckNormalizedData(self):
        isNormalized = True
        for d in range(self.d_CATE):
            if isNormalized == False: break
            for dvalue in range(self.D_CATE[d]):
                if dvalue not in self.uniques_CATE[d]:
                    isNormalized = False; break
        #print("\n\n\n ",isNormalized,"\n\n\n\n\n")
        if isNormalized== False:
            self.normalizer = CategoricalDatasetNormalization(self.X_CATE)
            self.X_CATE = self.normalizer.Normalize(self.X_CATE)
        else: self.normalizer = None

    def DoCluster(self,n_group=2):
        n = len(self.X_NUM)
        d_CATE = self.X_CATE.shape[1]
        d_NUM = self.X_NUM.shape[1]

        self.AddVariableToPrint("n_group",n_group)
        self.k = k = n_clusters = self.k
        self.farest_point = np.zeros((self.k,2))

        self.name = "LSHkRepresentatives"
        
        
        
        


        all_labels = []
        all_costs = []
        start_time = timeit.default_timer()
        #self.n_init=1; # Force set self.n_init
        for init_no in range(self.n_init):
            self.random_state = check_random_state(None)
            membship = np.zeros((k, n), dtype=np.uint8)
            labels_matrix = np.empty(n, dtype=np.uint16)
            
            for i in range(n): labels_matrix[i] = 65535
            
            means_count = np.zeros((self.k))
            means_sum = np.zeros((self.k, self.d_NUM))
            means = np.zeros((self.k, self.d_NUM))

            representatives_count = [[[0 for i in range(self.D_CATE[j])] for j in range(d_CATE)]for ki in range(k)]
            representatives_sum = [0 for ki in range(k)]
            last_cost = float('inf')

            representatives = [[[0 for i in range(self.D_CATE[j])] for j in range(d_CATE)] for ki in range(k)]
            

            buckets = [(k,len(self.lsh.hashTable[k])) for k in self.lsh.hashTable.keys()]
            buckets2 = sorted(buckets, key=lambda x: -x[-1])
            buckets_map_to_centroids = {}
            
            masterkeys=[]
            n_group = int(self.k/n_group)
            self.near_clusters = [[] for j in range(self.k)]
            self.lsh_group = [0 for i in range(self.n)]
            for i in range( self.k - len(buckets2)):
                buckets2.append((0,0))
            for i in range(self.k):
                masterkeys.append(buckets2[i][0])
                buckets_map_to_centroids[buckets2[i][0]] = i

            dist_from_master_to_other = [[ self.lsh.hammingDistance(keymaster,key) for key in self.lsh.hashTable.keys()] for keymaster in masterkeys  ]
            dist_from_master_to_master = [[ self.lsh.hammingDistance(keymaster,key) for key in masterkeys] for keymaster in masterkeys  ]

            count_remains = [n_group+1 for i in range(self.k) ]
            ddd= init_no%self.k; 
            #ddd= random.randint(0, self.k)
            ddd_end = ddd+self.k
            for ki_ in range(ddd,ddd_end):
                ki = ki_%self.k
                self.near_clusters[ki] = np.argsort(dist_from_master_to_master[ki])[0:n_group+1]
                if ki not in self.near_clusters[ki]:
                    self.near_clusters[ki][n_group]=ki
                for i in self.near_clusters[ki]:
                    count_remains[i] -=1
                    if count_remains[i] <=0:
                        for ki2 in range(ki,self.k):
                            dist_from_master_to_master[ki2][i] = float('inf')

            for key_id, key in enumerate(self.lsh.hashTable.keys()):
                nearest_key=-1
                nearest_dist = float('inf')
                for keymaster_id, value in enumerate(masterkeys):
                    d_temp = dist_from_master_to_other[keymaster_id][key_id]
                    if d_temp < nearest_dist: nearest_dist= d_temp; nearest_key = value;
                ki = buckets_map_to_centroids[nearest_key]
                
                for i in self.lsh.hashTable[key]:
                    labels_matrix[i] = ki
                    membship[ki][i]=1
                    representatives_sum[ki]+=1
                    for ii, val in enumerate(self.X_CATE[i]):
                        representatives_count[ki][ii][val]+=1
                    self.lsh_group[i] = ki

                    means_count[ki] += 1
                    for di in range(self.d_NUM):
                        means_sum[ki][di] += self.X_NUM[i][di]



            self.CheckEmptyClusters(representatives, self.X_CATE,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_count,means)
            self.UpdateRepresentatives(representatives,representatives_sum,representatives_count,means_count,means_sum,means) 
            for i in range(self.n_iter):
                self.iter = i
                cost , move, count_empty = self.UpdateLabels(representatives, self.X_CATE,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_sum,means)
                self.UpdateRepresentatives(representatives,representatives_sum,representatives_count,means_count,means_sum,means ) 
                if last_cost == cost and move==0: 
                    last_cost = self.UpdateLabelsLast(representatives, self.X_CATE,representatives_sum, representatives_count,membship,labels_matrix,means_count,means_sum,means)
                    break 
                last_cost = cost

            labels = self.GetLabels(membship)
            all_costs.append(cost)
            all_labels.append(labels)
            
        best = np.argmin(all_costs)
        labels = all_labels[best]
        self.time_score = (timeit.default_timer() - start_time)/ self.n_init
        self.labels = labels
        print("LSH time:", self.time_lsh ,"Score: ", all_costs[best] , " Time:", self.time_score)
        self.representatives= representatives
        self.means = means
        return self.labels
        
    

    def predict(self,x):    
        if len(x.shape) ==1:
            cateValues = x[self.indexes_CATE]
            if self.normalizer!=None:
                cateValues = self.normalizer.Normalize(cateValues)
            dist_matrix = self.DistancePrototypestoAPoints(self.representatives,self.means,cateValues ,x[self.indexes_NUM].astype(float)  )
            return dist_matrix[0]
        else:
            out = []
            for xi in x:
                out.append(self.predict(xi))
            return np.array(out)





import os
import os.path
import sys
from sys import platform
import numpy as np
from sklearn.utils.validation import check_array
from sklearn.metrics.cluster import adjusted_rand_score
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.metrics.cluster import adjusted_mutual_info_score
from sklearn.metrics.cluster import homogeneity_score
from sklearn.metrics.cluster import silhouette_score
import timeit
import csv
from csv import writer
import random
import math
def nCr(n,r):
    f = math.factorial
    return f(n) // f(r) // f(n-r)
class ClusteringAlgorithm:
    ALGORITHM_LIST = ['kModes','kRepresentatives']
    # def __init__(self, X, y,n_init=5,n_clusters=-1,max_iter=100,dbname='dpname',verbose =0,random_state = None):
    #     self.random_state= random_state
    #     self.verbose = verbose 
    #     self.measurename = 'None'
    #     self.dicts = [];self.dicts2 = []
    #     self.iter=-1
    #     self.dbname = dbname
    #     self.time_lsh=-1
    #     self.X = X
    #     self.y = y
    #     self.n = len(self.X)
    #     self.d = len(self.X[0])
    #     self.k = n_clusters if n_clusters > 0 else len(np.unique(y))
    #     self.n_init = n_init
    #     self.n_iter = max_iter
    #     self.scorebest = -2
    def __init__(self,n_clusters,n_init=5,max_iter=100,dbname="None",verbose =0,random_state = None):
        self.y = None
        self.random_state= random_state
        self.verbose = verbose 
        self.measurename = 'None'
        self.dicts = [];self.dicts2 = []
        self.iter=-1
        self.dbname = dbname
        self.time_lsh=-1
        
        self.k = n_clusters if n_clusters > 0 else len(np.unique(y))
        self.n_init = n_init
        self.n_iter = max_iter
        self.scorebest = -2

    #attributeMasks: masking the types of attributes: 0 -> Categorical ; 1 -> Numeric  : Example: [0,1,0,1] means the 1st and 3rd attributes are the categircal attribues while 2st and 4rd attributes are the numeric attribues
    def fit(self,X,attributeMasks=None, categorical_weight=1, numerical_weight=1):
        """
X: dataset
attributeMasks: masking the types of attributes: 
    0 -> Categorical ; 1 -> Numeric  : 
    Example: [0,1,0,1] means the 1st and 3rd attributes are the categircal attribues while 2st and 4rd attributes are the numeric attribues
categorical_weight: weight of categorical attributes
numerical_weight: weight of numericcal attributes

        """
        self.weight_CATE = categorical_weight
        self.weight_NUM = numerical_weight
        if type(attributeMasks) == list:
            attributeMasks = np.array(attributeMasks)
        if attributeMasks is None:
            attributeMasks = np.zeros((len(X[0])))
        self.attributeMasks = attributeMasks

        self.indexes_CATE =indexes_CATE= np.argwhere(attributeMasks==0)[:,0]
        self.indexes_NUM = indexes_NUM  = np.argwhere(attributeMasks>0)[:,0]

        
        self.X_CATE = X[:,indexes_CATE]
        self.X_NUM = X[:,indexes_NUM].astype(float)

        
        self.d_NUM = len(self.X_NUM[0])

        self.n = len(self.X_CATE)
        self.d_CATE = len(self.X_CATE[0])
        self.uniques_CATE =  []
        for i in range(self.d_CATE):
            unique = np.unique(self.X_CATE[:,i])
            self.uniques_CATE.append(unique)
        self.D_CATE  = [len(self.uniques_CATE[i]) for i in range(self.d_CATE) ]
        


        self.SetupLSH()
        self.DoCluster()
        return self.labels
    def SetupMeasure(self, classname):
        self.measurename = classname
        module = __import__(classname, globals(), locals(), ['object'])
        class_ = getattr(module, classname)
        self.measure = class_()
        self.measure.setUp(self.X, self.y)
    def Overlap(self,x,y):
        n = len(x)
        sum =0
        for i in range(n):
            if x[i] != y[i]: sum +=1
        return sum
    def DoCluster(self):
        print("Do something")
        return -1
    def _labels_cost(self,X, centroids, dissim, membship=None):
        X = check_array(X)
        n_points = X.shape[0]
        cost = 0.
        labels = np.empty(n_points, dtype=np.uint16)
        for ipoint, curpoint in enumerate(X):
            diss = self.ComputeDistances(centroids, curpoint)
            clust = np.argmin(diss)
            labels[ipoint] = clust
            cost += diss[clust]
        return labels, cost
    def ComputeDistances(self, X, mode):
        return [ self.measure.calculate(i, mode ) for i in X ]
    def CheckCLusteringPurityByHeuristic(self,labels_,km_labels_):
        #start = timeit.default_timer()
        unique_ = np.unique(labels_)
        n_clusters = len(unique_)
        matching_matrix = [-1 for i in range (n_clusters)]
        n = len(labels_)
        n_range = range(n)
        #Computer matching matrix
        count_item =0
        for i in range(n_clusters):
            max_count =-1
            max_index = 0
            for j in range(n_clusters):
                if j in matching_matrix:
                    continue
                count = sum([labels_[k] == i and km_labels_[k] == j for k in n_range ])
                if count > max_count:
                    max_count = count
                    max_index = j
            matching_matrix[i] = max_index
            count_item =count_item+ max_count
        #Compute score
        score2 = count_item/len(labels_)
        return score2
    def AcPrRc(self,A,B):
        #B = np.array([0,0,0,0,0,1,2,2,   0,1,1,1,1,   1,2,2,2 ])
        #B = np.array([1,1,1,1,1,2,0,0,   1,2,2,2,2,   2,0,0,0 ])

        k = len(np.unique(A))
        n= len(A)
        MAP =[0,1,2]

        clustersA = [np.where(A==i)[0] for i in range(k)]
        clustersB = [np.where(B==i)[0] for i in range(k)]

        TP_FP=0
        for i in range(k):
            if len(clustersB[i]) > 1: 
                TP_FP+=  nCr(len(clustersB[i]),2)
        TP=0
        for i in range(k):
            for j in range(k):
                num = sum(A[clustersB[i]]==j)
                if num > 1:
                    TP+= nCr(num,2)

        FP = TP_FP-TP


        TN_FN = int(n*(n-1)/2) - TP_FP
        FN=0

        for i in range(k):
            for j in range(k):
                num_ = sum(A[clustersB[i]]==j)
                sum_=0
                for i2 in range(i+1,k):
                    sum_ += sum(A[clustersB[i2]]==j)
                FN+=num_*sum_
        TN = TN_FN - FN
        #print('TP=',TP,'TP_FP=',TP_FP ,'TN_FN=',TN_FN,'FN=',FN)
        PR = TP/(TP+FP)
        RC = TP/(TP+FN)
        AC = (TP+TN)/(TP+TN+FP+FN)
        #print('PR=',PR,'RC=',RC,'AC=',AC )
        return AC,PR,RC
    def CalcScore(self,y, verbose=True):
        self.y = y
        starttime = timeit.default_timer()
        s="";
        if self.n*self.k <= 8000000: 
            self.purity_score = self.CheckCLusteringPurityByHeuristic(self.y, self.labels)
        else: self.purity_score =-2
        s+= str(timeit.default_timer()-starttime)+"|";  starttime = timeit.default_timer()
        self.NMI_score = normalized_mutual_info_score(self.y,self.labels) #tulti.CheckClusteringNMI(self.y, self.labels)
        s+= str(timeit.default_timer()-starttime)+"|";  starttime = timeit.default_timer()
        self.ARI_score = adjusted_rand_score(self.labels,self.y) # tulti.CheckClusteringARI(self.y, self.labels)
        s+= str(timeit.default_timer()-starttime)+"|";  starttime = timeit.default_timer()
        self.AMI_score = adjusted_mutual_info_score(self.labels,self.y)
        s+= str(timeit.default_timer()-starttime)+"|";  starttime = timeit.default_timer()
        self.HOMO_score = homogeneity_score(self.labels,self.y)
        s+= str(timeit.default_timer()-starttime)+"|";  starttime = timeit.default_timer()
        if self.n*self.k <= 800000:
            try: 
                self.SILHOUETTE_score = silhouette_score(self.X, self.labels, metric= self.Overlap)
            except:
                self.SILHOUETTE_score=-1
        else: self.SILHOUETTE_score=-2

        s+= str(timeit.default_timer()-starttime)+"|";  starttime = timeit.default_timer()
        if self.n*self.k <= 8000000: 
            self.Ac_score, self.Pr_score,self.Rc_score =  self.AcPrRc(self.y, self.labels)
        else: self.Ac_score =  self.Pr_score = self.Rc_score = -2  

        s+= str(timeit.default_timer()-starttime)+"|";  starttime = timeit.default_timer()
        self.AddVariableToPrint("Scoringtime",s )
        self.WriteResultToCSV()
        if verbose: print("Purity:", "%.2f" % self.purity_score,"NMI:", "%.2f" %self.NMI_score,"ARI:", "%.2f" %self.ARI_score,"Sil: ", "%.2f" %self.SILHOUETTE_score,"Acc:", "%.2f" %self.Ac_score,
                          "Recall:", "%.2f" %self.Rc_score,"Precision:", "%.2f" %self.Pr_score)
        return (self.purity_score,self.NMI_score,self.ARI_score,self.AMI_score,self.HOMO_score,self.SILHOUETTE_score,self.time_score, self.time_lsh,
                self.Ac_score, self.Pr_score, self.Rc_score)

    def append_list_as_row(self,file_name, list_of_elem):
        with open(file_name, 'a+', newline='') as write_obj:
            csv_writer = writer(write_obj)
            csv_writer.writerow(list_of_elem)
    def AddVariableToPrint(self,name,val):
        self.dicts2.append((name,val ))

    def WriteResultToCSV(self,file=''):
        return 0
        if not os.path.exists(TDef.folder):
            os.makedirs(TDef.folder)
        if file=='':
            file = TDef.folder+ '/' + self.name + TDef.fname + ".csv" 
        
        self.dbname = self.dbname.replace("_c","").replace(".csv","").capitalize()
        self.dicts.append(('dbname',self.dbname ))
        self.dicts.append(('n',self.n ))
        self.dicts.append(('d',self.d ))
        self.dicts.append(('k',self.k ))
        self.dicts.append(('range','-1' ))
        self.dicts.append(('sigma_ratio',-1 ))
        self.dicts.append(('Measure',  self.measurename))
        self.dicts.append(('n_init',self.n_init ))
        self.dicts.append(('n_iter',self.n_iter ))
        self.dicts.append(('iter',self.iter ))
        self.dicts.append(('Purity',self.purity_score ))
        self.dicts.append(('NMI',self.NMI_score ))
        self.dicts.append(('ARI',self.ARI_score ))
        self.dicts.append(('AMI',self.AMI_score ))
        self.dicts.append(('Homogeneity',self.HOMO_score ))
        self.dicts.append(('Silhouette',self.SILHOUETTE_score ))
        self.dicts.append(('Accuracy',self.Ac_score ))
        self.dicts.append(('Precision',self.Pr_score ))
        self.dicts.append(('Recall',self.Rc_score ))
        self.dicts.append(('Time',self.time_score ))
        self.dicts.append(('LSH_time',self.time_lsh ))
        self.dicts.append(('Score',self.scorebest ))
        
        dicts = self.dicts+self.dicts2;
        try:
            if os.path.isfile(file)==False:
                colnames = [i[0] for i in dicts]
                self.append_list_as_row(file,colnames)
            vals = [i[1] for i in dicts]
            self.append_list_as_row(file,vals)
        except Exception  as ex:
            #self.exe(file,ex)
            print('Cannot write to file ', file ,'', ex);
            self.WriteResultToCSV(file + str(random.randint(0,1000000)) + '.csv')

    def AddValuesToMyTable(self, mytable, is_first=False,dbname=''):
        if is_first:
            mytable.AddValue("Purity", 'Dataset' ,dbname )
            mytable.AddValue("NMI", 'Dataset',dbname )
            mytable.AddValue("ARI", 'Dataset',dbname )
            mytable.AddValue("AMI", 'Dataset',dbname )
            mytable.AddValue("Homogeneity", 'Dataset',dbname )
            mytable.AddValue("Silhouette", 'Dataset',dbname )
            mytable.AddValue("iter", 'Dataset', dbname)
            mytable.AddValue("Precision", 'Dataset',dbname )
            mytable.AddValue("Recall", 'Dataset',dbname )
            mytable.AddValue("Accuracy", 'Dataset',dbname )
            mytable.AddValue("LSH_time", 'Dataset',dbname )
            mytable.AddValue("Time", 'Dataset',dbname )
            mytable.AddValue("Time", 'Dataset',dbname )
            

        mytable.AddValue("Purity", self.measurename ,self.purity_score )
        mytable.AddValue("NMI", self.measurename,self.NMI_score )
        mytable.AddValue("ARI", self.measurename,self.ARI_score )
        mytable.AddValue("AMI", self.measurename,self.AMI_score )
        mytable.AddValue("Homogeneity", self.measurename,self.HOMO_score )
        mytable.AddValue("Silhouette", self.measurename,self.SILHOUETTE_score )
        mytable.AddValue("iter", self.measurename, self.iter )
        mytable.AddValue("Precision", self.measurename,self.Pr_score )
        mytable.AddValue("Recall", self.measurename,self.Rc_score )
        mytable.AddValue("Accuracy", self.measurename,self.Ac_score )
        mytable.AddValue("LSH_time", self.measurename,self.time_lsh )
        mytable.AddValue("Time", self.measurename,self.time_score )



import os
import os.path
import sys
from sys import platform

import numpy as np
import pandas as pd
#from kmodes_lib import KModes
from collections import defaultdict
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_array
import timeit
#from kmodes.util import get_max_value_key, encode_features, get_unique_rows, \
#    decode_centroids, pandas_to_numpy

from .ClusteringAlgorithm import ClusteringAlgorithm
from sklearn.metrics.cluster import adjusted_rand_score
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.metrics.cluster import adjusted_mutual_info_score
from sklearn.metrics.cluster import homogeneity_score
import random
from .LSH import LSH 
import multiprocessing as mp
import math
from collections import defaultdict

class CategoricalDatasetNormalization:
    def __init__(self,X):
        self.d = X.shape[1]
        self.uniques = []
        self.maps = []
        for di in range(self.d):
            x = X[:,di]
            unique = np.unique(x)
            mapi={}
            for i in range(len(unique)):
                mapi[unique[i]] = i
            self.maps.append(mapi)
            self.uniques.append(unique)
        
    
    def Normalize(self,X):
        if len(X.shape) ==2:
            if X.shape[1] != self.d: raise Exception(f"ERROR, Dimension of the data should be {self.d} (must same as size of trained dataset)")
            Xnormalized = np.zeros((X.shape),dtype=int)
            for i in range(X.shape[0]):
                for j in range(X.shape[1]):
                    Xnormalized[i,j] = self.maps[j][X[i,j]]
            return Xnormalized
        else:
            if X.shape[0] != self.d: raise Exception(f"ERROR, Dimension of the data should be {self.d} (must same as size of trained dataset)")
            Xnormalized = np.zeros((X.shape),dtype=int)
            for j in range(X.shape[0]):
                Xnormalized[j] = self.maps[j][str(X[j])]
            return Xnormalized
        
class LSHkRepresentatives(ClusteringAlgorithm):

    def SetupLSH(self, hbits=-1,k=-1,measure='DILCA' ):
        self.d = self.d_CATE
        self.D = self.D_CATE
        self.X = self.X_CATE
        self.uniques = self.uniques_CATE
        if self.d_NUM > 0:
            print("\n\n YOU ARE SETTING SOME NUMERIC ATTRIBUTES IN YOUR DATA, SO PLEASE USE LSHkPrototypes INSTEAD OF LSHkRepresentatives \n\n")
        
        # self.n = n = self.X.shape[0]
        # self.d = d = self.X.shape[1]
        

        
        self.CheckNormalizedData()

        #hbits = 2*math.ceil(math.log(self.k)/math.log(2))
        start = timeit.default_timer()
        self.lsh = LSH(self.X,self.y,measure=measure,hbits=hbits)
        self.lsh.DoHash()
        #self.lsh.CorrectSingletonBucket()
        
        self.time_lsh = timeit.default_timer() - start
        self.AddVariableToPrint("Time_lsh",self.time_lsh)
        return self.time_lsh
    def SetupMeasure(self, classname):
        self.measurename = classname
        module = __import__(classname, globals(), locals(), ['object'])
        class_ = getattr(module, classname)
        self.measure = class_()
        self.measure.setUp(self.X, self.y)
    def test(self):
        print("a234 " + str(self.k))
    def Distance(self,representative, point):
        sum=0
        for i in range (self.d):
            sum = sum + representative[i][point[i]]
        return self.d - sum
    
    def MovePoint(self, point_id, from_id, to_id ,representatives_count, representatives_sum,membship, curpoint,labels_matrix):
        labels_matrix[point_id] = to_id
        membship[to_id, point_id] = 1
        membship[from_id, point_id] = 0
        representatives_sum[to_id]+=1
        representatives_sum[from_id]-=1 
        for ii, val in enumerate(curpoint):
            representatives_count[to_id][ii][val]+=1
            representatives_count[from_id][ii][val]-=1
    def CheckEmptyClusters(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        move =0
        big_cluster_id = -1
        for ki in range(self.k):
            if representatives_sum[ki] ==0 :
                #print("EMPTY: ", ki)
                big_cluster_id = np.argmax([sum(mem_) for mem_ in membship])
                choices = [i for i in range(self.n) if membship[big_cluster_id][i] == 1 ]
                #rindx = self.random_state.choice(choices)
                rindx = int(self.farest_point[big_cluster_id][0])
                self.MovePoint(rindx, big_cluster_id,ki, representatives_count, representatives_sum,membship,self.X[rindx],labels_matrix  )
                move +=1

        return move
    def InitClusters(self,representatives,representatives_sum,representatives_count):
        for ki in range(self.k):
            for i in range(self.d):
                sum_ = 0
                for j in range(self.D[i]): sum_ = sum_ + representatives[ki][i][j]
                for j in range(self.D[i]): representatives[ki][i][j] = representatives[ki][i][j]/sum_;

    def DistanceRepresentativestoAPoints(self,representatives, point):
        dist_matrix = [self.Distance(c, point) for c in representatives]
        representative_id = np.argmin(dist_matrix)
        return representative_id, dist_matrix[representative_id]

    def UpdateLabelsInit(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        cost = 0
        move = 0
        self.preferList= defaultdict(set)
        for ipoint, curpoint in enumerate(X):
            representative_id,tmp = self.DistanceRepresentativestoAPoints(representatives, curpoint)
            cost += tmp
            labels_matrix[ipoint] = representative_id
            membship[representative_id, ipoint] = 1
            representatives_sum[representative_id]+=1
            for ii, val in enumerate(curpoint):
                representatives_count[representative_id][ii][val]+=1
            self.preferList[self.lsh.hash_values[ipoint]].add(labels_matrix[ipoint])
        self.CheckEmptyClusters(representatives, X,representatives_sum, representatives_count,membship,labels_matrix)
        self.dist_matrix_tmp = [1000000000 for i in range(self.k)]
        return cost ,move, 0
    def UpdateLabelsLast(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        cost = 0
        self.preferList= defaultdict(set)


        for ipoint, curpoint in enumerate(X):
            representative_id,tmp = self.DistanceRepresentativestoAPoints(representatives, curpoint)
            cost += tmp
            membship[representative_id, ipoint] = 1
        return cost ,0, 0

    def DistanceRepresentativestoAPoints_LSH(self,item_id, point,labels_matrix,representatives):
        myset = self.preferList[self.lsh.hash_values[item_id]]
        dist_min = 1000000000
        dist_index =-1
        for i in myset:
            dist = self.Distance(representatives[i], point)
            if dist_min > dist:
                dist_min = dist 
                dist_index = i
        return dist_index, dist_min 
    def DistanceRepresentativestoAPoints_LSH2(self,item_id, point,labels_matrix,representatives):
        #myset = self.preferList[self.lsh.hash_values[item_id]]
        #myset = self.near_clusters[self.lsh_group[item_id]]
        myset = self.near_clusters[labels_matrix[item_id]]

        dist_min = 1000000000
        dist_index =-1
        for i in myset:
            dist = self.Distance(representatives[i], point)
            if dist_min > dist:
                dist_min = dist 
                dist_index = i
        return dist_index, dist_min 

    def UpdateLabels(self,representatives, X,representatives_sum, representatives_count,membship,labels_matrix):
        cost = 0
        move = 0
        for i in range(self.k):
            self.farest_point[i][1] = 0

        for ipoint, curpoint in enumerate(X):
            #representative_id,tmp = self.DistanceRepresentativestoAPoints_LSH(ipoint, curpoint,labels_matrix,representatives)
            representative_id,tmp = self.DistanceRepresentativestoAPoints(representatives, curpoint)
            #representative_id,tmp = self.DistanceRepresentativestoAPoints_LSH2(ipoint, curpoint,labels_matrix,representatives)
            if tmp > self.farest_point[representative_id][1]:
                self.farest_point[representative_id][1] = tmp
                self.farest_point[representative_id][0] = ipoint
            cost += tmp
            
            if membship[representative_id, ipoint]: continue
            old_clust = labels_matrix[ipoint]
            self.MovePoint(ipoint, old_clust,representative_id, representatives_count, representatives_sum,membship,curpoint,labels_matrix  )
            move +=1
        #Check empty clusters
        move  += self.CheckEmptyClusters(representatives, X,representatives_sum, representatives_count,membship,labels_matrix)
        return cost ,move, 0

    def UpdateRepresentatives(self,representatives,representatives_sum,representatives_count ) :  
        for ki in range(self.k):
            for di in range(self.d):
                    for vj in range(self.D[di]):
                        representatives[ki][di][vj] =  representatives_count[ki][di][vj]/representatives_sum[ki]
        return 0
    def GetLabels(self, membship):
        labels = np.empty(self.n, dtype=np.uint16)
        for ki in range(self.k):
            for i in range(self.n):
                if membship[ki][i]:
                    labels[i] = ki
        return labels
    def CheckNormalizedData(self):
        isNormalized = True
        for d in range(self.d):
            if isNormalized == False: break
            for dvalue in range(self.D[d]):
                if dvalue not in self.uniques[d]:
                    isNormalized = False; break
        #print("\n\n\n ",isNormalized,"\n\n\n\n\n")
        if isNormalized== False:
            self.normalizer = CategoricalDatasetNormalization(self.X)
            self.X = self.normalizer.Normalize(self.X)
        else: self.normalizer = None

    def DoCluster(self,n_group=2):
        n = len(self.X)
        d = self.X.shape[1]
        self.AddVariableToPrint("n_group",n_group)
        self.k = k = n_clusters = self.k
        self.farest_point = np.zeros((self.k,2))

        self.name = "LSHkRepresentatives"
        
        
        
        


        all_labels = []
        all_costs = []
        start_time = timeit.default_timer()
        #self.n_init=1; # Force set self.n_init
        for init_no in range(self.n_init):
            self.random_state = check_random_state(None)
            membship = np.zeros((k, n), dtype=np.uint8)
            labels_matrix = np.empty(n, dtype=np.uint16)
            
            for i in range(n): labels_matrix[i] = 65535
            representatives_count = [[[0 for i in range(self.D[j])] for j in range(d)]for ki in range(k)]
            representatives_sum = [0 for ki in range(k)]
            last_cost = float('inf')

            representatives = [[[0 for i in range(self.D[j])] for j in range(d)] for ki in range(k)]
            buckets = [(k,len(self.lsh.hashTable[k])) for k in self.lsh.hashTable.keys()]
            buckets2 = sorted(buckets, key=lambda x: -x[-1])
            buckets_map_to_centroids = {}
            
            masterkeys=[]
            n_group = int(self.k/n_group)
            self.near_clusters = [[] for j in range(self.k)]
            self.lsh_group = [0 for i in range(self.n)]
            for i in range( self.k - len(buckets2)):
                buckets2.append((0,0))
            for i in range(self.k):
                masterkeys.append(buckets2[i][0])
                buckets_map_to_centroids[buckets2[i][0]] = i

            dist_from_master_to_other = [[ self.lsh.hammingDistance(keymaster,key) for key in self.lsh.hashTable.keys()] for keymaster in masterkeys  ]
            dist_from_master_to_master = [[ self.lsh.hammingDistance(keymaster,key) for key in masterkeys] for keymaster in masterkeys  ]

            count_remains = [n_group+1 for i in range(self.k) ]
            ddd= init_no%self.k; 
            #ddd= random.randint(0, self.k)
            ddd_end = ddd+self.k
            for ki_ in range(ddd,ddd_end):
                ki = ki_%self.k
                self.near_clusters[ki] = np.argsort(dist_from_master_to_master[ki])[0:n_group+1]
                if ki not in self.near_clusters[ki]:
                    self.near_clusters[ki][n_group]=ki
                for i in self.near_clusters[ki]:
                    count_remains[i] -=1
                    if count_remains[i] <=0:
                        for ki2 in range(ki,self.k):
                            dist_from_master_to_master[ki2][i] = float('inf')

            for key_id, key in enumerate(self.lsh.hashTable.keys()):
                nearest_key=-1
                nearest_dist = float('inf')
                for keymaster_id, value in enumerate(masterkeys):
                    d_temp = dist_from_master_to_other[keymaster_id][key_id]
                    if d_temp < nearest_dist: nearest_dist= d_temp; nearest_key = value;
                ki = buckets_map_to_centroids[nearest_key]
                #ki= random.randint(0, self.k-1 )
                for i in self.lsh.hashTable[key]:
                    labels_matrix[i] = ki
                    membship[ki][i]=1
                    representatives_sum[ki]+=1
                    for ii, val in enumerate(self.X[i]):
                        representatives_count[ki][ii][val]+=1
                    self.lsh_group[i] = ki

            self.CheckEmptyClusters(representatives, self.X,representatives_sum, representatives_count,membship,labels_matrix)
            self.UpdateRepresentatives(representatives,representatives_sum,representatives_count ) 
            for i in range(self.n_iter):
                self.iter = i
                cost , move, count_empty = self.UpdateLabels(representatives, self.X,representatives_sum, representatives_count,membship,labels_matrix)
                self.UpdateRepresentatives(representatives,representatives_sum,representatives_count ) 
                if last_cost == cost and move==0: 
                    last_cost = self.UpdateLabelsLast(representatives, self.X,representatives_sum, representatives_count,membship,labels_matrix)
                    #print("last_cost=", last_cost, "last_cost2=",last_cost2)
                    break 
                last_cost = cost
                #print ("Iter: ", i , " Cost:", cost, "Move:", move)
            labels = self.GetLabels(membship)
            all_costs.append(cost)
            all_labels.append(labels)
            
        best = np.argmin(all_costs)
        labels = all_labels[best]
        self.time_score = (timeit.default_timer() - start_time)/ self.n_init
        self.labels = labels
        print("LSH time:", self.time_lsh ,"Score: ", all_costs[best] , " Time:", self.time_score)
        self.representatives= representatives
        return self.labels
        # Update representives
    def predict(self,x):
        if self.normalizer!=None:
            x = self.normalizer.Normalize(x)

        if len(x.shape) ==1:
            dist_matrix = self.DistanceRepresentativestoAPoints(self.representatives, x)
            return dist_matrix[0]
        else:
            out = []
            for xi in x:
                out.append(self.DistanceRepresentativestoAPoints(self.representatives, xi)[0])
            return np.array(out)


     
def Test_Simple():
    DB = tulti.LoadSynthesisData(n=128,d=16,k=8,sigma_rate=0.1); 
    MeasureManager.CURRENT_DATASET = DB['name']
    MeasureManager.CURRENT_MEASURE = 'DILCA'

    print("\n\n############## LSHkRepresentatives ###################")
    lshkrepresentatives = LSHkRepresentatives(DB['DB'],DB['labels_'] )
    lshkrepresentatives.SetupMeasure(MeasureManager.CURRENT_MEASURE)
    lshkrepresentatives.SetupLSH(measure=MeasureManager.CURRENT_MEASURE)
    lshkrepresentatives.DoCluster()
    lshkrepresentatives.CalcScore()

    print("\n\n############## KMODES ###################")
    kmodes = kModes(DB['DB'],DB['labels_'] )
    kmodes.SetupMeasure(MeasureManager.CURRENT_MEASURE)
    kmodes.DoCluster()
    kmodes.CalcScore()

    print("\n\n############## kRepresentatives ###################")
    kmodes = kRepresentatives(DB['DB'],DB['labels_'] )
    kmodes.SetupMeasure(MeasureManager.CURRENT_MEASURE)
    kmodes.DoCluster()
    kmodes.CalcScore()

def Test(): 
    MeasureManager.CURRENT_DATASET = 'balance-scale.csv'
    MeasureManager.CURRENT_MEASURE = 'Overlap'
    if TDef.data!='': MeasureManager.CURRENT_DATASET = TDef.data
    if TDef.measure!='': MeasureManager.CURRENT_MEASURE = TDef.measure
    if TDef.test_type == 'syn':
        DB = tulti.LoadSynthesisData(TDef.n,  TDef.d, TDef.k)
        MeasureManager.CURRENT_DATASET= DB['name']
    else:
        DB = tulti.LoadRealData(MeasureManager.CURRENT_DATASET)
    print("\n\n############## LSHkRepresentatives ###################")
    lshkrepresentatives = LSHkRepresentatives(DB['DB'],DB['labels_'] ,dbname=MeasureManager.CURRENT_DATASET ,k=TDef.k)
    lshkrepresentatives.SetupMeasure(MeasureManager.CURRENT_MEASURE)
    lshkrepresentatives.SetupLSH(measure=MeasureManager.CURRENT_MEASURE)
    lshkrepresentatives.DoCluster()
    lshkrepresentatives.CalcScore()


def TestDatasets(): 
    for dbname in MeasureManager.DATASET_LIST:
        DB = tulti.LoadRealData(dbname)
        MeasureManager.CURRENT_DATASET = dbname
        MeasureManager.CURRENT_MEASURE = 'Overlap'
        print("\n\n############## LSHkRepresentatives ###################")
        alo = LSHkRepresentatives(DB['DB'],DB['labels_'],dbname=MeasureManager.CURRENT_DATASET )
        alo.SetupMeasure(MeasureManager.CURRENT_MEASURE)
        alo.SetupLSH(measure=MeasureManager.CURRENT_MEASURE)
        alo.DoCluster()
        alo.CalcScore()
def TestMeasures():
    table = MyTable()
    for measure in MeasureManager.MEASURE_LIST:
        MeasureManager.CURRENT_MEASURE = measure
        for dataset in MeasureManager.DATASET_LIST:
            MeasureManager.CURRENT_DATASET=dataset
            DB = tulti.LoadRealData(MeasureManager.CURRENT_DATASET)
            alo = LSHkRepresentatives(DB['DB'],DB['labels_'],dbname=MeasureManager.CURRENT_DATASET )
            alo.SetupMeasure(MeasureManager.CURRENT_MEASURE)
            alo.SetupLSH(measure=MeasureManager.CURRENT_MEASURE)
            alo.DoCluster()
            alo.CalcScore()
            alo.AddValuesToMyTable(table,measure == MeasureManager.MEASURE_LIST[0],dataset)
            table.SaveToExcelFolder("RESULT/r20200916",alo.name,[i+1 for i in range(0,1000)])

if __name__ == "__main__":
    X = np.array([[0,0,0],[0,1,1],[0,0,0],[1,0,1],[2,2,2],[2,3,2],[2,3,2]])
    kreps = LSHkRepresentatives(n_clusters=2,n_init=5) 
    kreps.fit(X)
    print()
    print(kreps.labels)
    print()

    print(kreps.predict(X[0]))
    print(kreps.predict(X[1]))
    print(kreps.predict(X[2]))
    print(kreps.predict(X[3]))
    print(kreps.predict(X[4]))
    print(kreps.predict(X[5]))
   



import os
import os.path
import sys
from sys import platform
sys.path.append(os.path.join(os.getcwd(), "Measures"))
import numpy as np
import pandas as pd

from collections import defaultdict
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_array
import timeit

import random
from .SimpleHashing import SimpleHashing
from collections import defaultdict
import statistics 
from collections import Counter
class LSH(SimpleHashing):
    def DoHash(self):

        self.measure.GeneratesimMatrix()
        self.GenerateSimilarityMatrix(self.measure.simMatrix)
        self.bit_indexes  = np.argpartition(self.cut_values_normal, self.hbits)[:self.hbits]   
        self.GenerateHashTable()
        return -1

    def GenerateHashTable(self):
        print("Generating LSH hash table: ", " hbits:", str(self.hbits) +'('+ str(2**self.hbits)+')', " k", self.k , " d", self.d , " n=",self.n )
        self.hash_values = [self.ComputeHashValue(x) for x in self.X]
        self.hashTable = defaultdict(list)
        for i in range(self.n):
            self.hashTable[self.hash_values[i]].append(i)

    def GetNeighborsbyBucket(self, item_id):
        return self.hashTable[self.hash_values[item_id]]

    def ComputeHashValue_Old(self,x): #NEW
        val=0
        for i in range(self.hbits):
            
            partitions = self.partitions[self.bit_indexes[i]]
            val <<=1
            if x[self.bit_indexes[i]] in partitions[1]:
                val+=1
        return val
    def ComputeHashValue(self,x): #NEW
        val=0
        for i in range(self.hbits):
            partitions = self.partitions[self.bit_indexes[i]]
            val <<=1
            if partitions is None:
                median = self.measure.medianValues[self.bit_indexes[i]]
                if x[self.bit_indexes[i]] < median:
                    val+=1
            else:
                
                if x[self.bit_indexes[i]] in partitions[1]:
                    val+=1
        return val

    def hammingDistance(self, x, y):
        ans = 0
        for i in range(31,-1,-1):
            b1= x>>i&1
            b2 = y>>i&1
            ans+= not(b1==b2)
        return ans
    def CorrectSingletonBucket(self):
        list_hash_1 = []
        for hashValue,itemList in self.hashTable.items():
            if(len(itemList)<=1): 
                list_hash_1.append((hashValue, itemList))

        for iters in list_hash_1:
            del self.hashTable[iters[0]]

        for iters in list_hash_1:
            closest_hash_value = -1
            closest_dist=1000000
            for hashValue,itemList in self.hashTable.items():
                temp = self.hammingDistance(iters[0],hashValue)
                if temp < closest_dist:
                    closest_hash_value = hashValue
                    closest_dist = temp
            for i in iters[1]:
                self.hash_values[i] =  closest_hash_value
                self.hashTable[closest_hash_value].append(i)
        print("LSH Merged ",len(list_hash_1),"/", len(self.hashTable) , " buckets!!!" )

    def TestHashTable(self):
        n = len(self.hashTable.items())
        num_0 = 2**self.hbits - len(self.hashTable.items());
        num_1 = 0;
        len_list=[]

        for hashValue,itemList in self.hashTable.items():
            if(len(itemList)==1): num_1+=1
            len_list.append(len(itemList))
        mean=np.mean(len_list)
        std_ = np.std(len_list)

        print( "Num bucket:",n," Num zero:", num_0, " Num 1:", num_1, " Mean:", mean, " Std:",std_)
        #Test within buckets
        sum_=0
        for hashValue,itemList in self.hashTable.items():
            labels = self.y[itemList]
            test_list = Counter(labels) 
            domiant_label = test_list.most_common(1)[0][0] 
            sum_+= sum(labels==domiant_label)
        return sum_/self.n
def main():
    MeasureManager.CURRENT_MEASURE = 'DILCA'
    DB = tulti.LoadSynthesisData(n=2048,d=10,k=20,sigma_rate=0.1);
    MeasureManager.CURRENT_DATASET = DB['name']
    hashing = LSH(DB['DB'],DB['labels_'],measure='DILCA')
    hashing.test()
    hashing.DoHash()
    score = hashing.TestHashTable()
    print('Score: ', score)
    hashing.CorrectSingletonBucket();
    score = hashing.TestHashTable()
    print('Score: ', score)
    asd = 123
if __name__ == "__main__":
    main()


import numpy as np

import math
import os
import os.path
import sys
from sys import platform
sys.path.append(os.path.join(os.getcwd(), "Measures"))
import networkx as nx
from collections import defaultdict
import math

class SimpleHashing:
    ALGORITHM_LIST = ['LSH']
    def __init__(self, X, y,n_init=10,hbits=-1,k=-1,measure='DILCA',attributeMasks=None):
        self.attributeMasks = attributeMasks
        self.X = X
        self.y = y
        self.n = len(self.X)
        self.d = len(self.X[0])
        self.D = D = [len(np.unique(self.X[:,i])) for i in range(self.d) ]
        self.k = k if k > 0 else len(np.unique(y))
        self.hbits = hbits if hbits > 0 else int( math.log(self.n*2/ len(np.unique(y)))/ math.log(2))
        
        while 2**(self.hbits-1) < self.k:
            self.hbits +=1
        if(self.hbits>=self.d) : self.hbits = self.d-1

        if 2**self.hbits <=self.k: print("WARNING: BAD HBITS: hbits=",self.hbits, " d=", self.d,  " nbucket=", 2**self.hbits," k=", self.k, ' n=',self.n)
        if n_init > 0: self.n_init = n_init
        self.SetupMeasure(measure)
    def SetupMeasure(self, classname):
        module = __import__(classname, globals(), locals(), ['object'])
        class_ = getattr(module, classname)
        self.measure = class_()
        self.measure.setUp(self.X, self.y)
    def SetupMeasure(self, classname):
        #module = __import__(classname, globals(), locals(), ['object'])
        #class_ = getattr(module, classname)
        if classname=='DILCA':
            from .DILCA import DILCA
            class_ = DILCA
        self.measure = class_()
        self.measure.setUp(self.X, self.y)
    def Overlap(self,x,y):
        n = len(x)
        sum =0
        for i in range(n):
            if x[i] != y[i]: sum +=1
        return sum
    def test(self):
        print("n=",self.n, " d=",self.d, ' hbit=', self.hbits, " k=",self.k , " measure=", self.measure)
    #def DoHash
    def GenerateSimilarityMatrix(self, simMatrix):
        self.partitions = [];
        self.cut_values =[]
        self.cut_values_normal =[]
        for di in range(self.d):
            if simMatrix[di] == None:
                self.partitions.append(None)
                self.cut_values.append(None)
                self.cut_values_normal.append(self.attributeMasks[di]*2)
                continue

            G = nx.Graph()
            for i in range(self.D[di]):
                matrix1D = [] # 1D array for 1 dimension
                for j in range(i+1,self.D[di]): 
                    #simMatrix[di][i][j]
                    G.add_edge(i,j,weight=simMatrix[di][i][j])
            if len(G.nodes)>1:
                ut_value, partition = nx.stoer_wagner(G)
                self.partitions.append(partition)
                self.cut_values.append(ut_value)
                self.cut_values_normal.append(ut_value/ self.D[di])
            else :
                self.partitions.append([[0],[0]])
                self.cut_values.append(10000000)
                self.cut_values_normal.append(10000000)


from .Measure import Measure
import math
from .MathUtils import *
from scipy.stats import entropy
import numpy as np
import timeit

class DILCA(Measure):
    def __init__(self):
        self.name = "DILCA"
    def ProconditionMatrixYX(self, Y, X,Y_unique, X_unique):
        lenData = len(Y)
        lenX = len(X_unique)
        lenY = len(Y_unique)
        HY = 0
        count_x = 0;
        count_y=0;
        MATRIX = [[0 for i in range(lenX)] for j in  range(lenY)]
        for x in range(lenX):
            for y in range(lenY):
                count_x = count_y =0
                for i in range(lenData):
                    if X[i] == x:
                        count_x = count_x+1
                        if(Y[i] == y):
                            count_y = count_y+1;
                MATRIX[y][x] = count_y/count_x if count_x > 0 else 0
                asd = 12332
        return MATRIX

    def ComputeEtropy(self,X):
        N = len(X);
        D = len(X[0]);
        context_array = [[] for _ in range(D)]
        self.contextMatrix = []
        self.conditionProMatrix= {}
        self.probabilityMatrix = [];
        #Compute cross contextMatrix 
        for i in range(D):
            Y_ = X[:,i]
            Y_unique, Y_freq = np.unique(Y_, return_counts=True)
            X_probability  = [ii/len(Y_) for ii in Y_freq]
            self.probabilityMatrix.append(X_probability)
            HY = entropy(X_probability,base=2)
            SUY_dict = {}
            for j in range(D):
                if i !=j:
                    X_ = X[:,j]
                    X_unique, X_freq = np.unique(X_, return_counts=True)
                    X_property = [ii/len(X_) for ii in X_freq]
                    HX = entropy(X_property,base=2)
                    conditionMatrix = self.ProconditionMatrixYX(Y_,X_,Y_unique, X_unique)
                    self.conditionProMatrix[(i,j)] = conditionMatrix
                    HYX = 0
                    for k in range(len(X_unique)):
                        sum_tmp =0;
                        for k2 in range(len(Y_unique)):
                            if conditionMatrix[k2][k] !=0:
                                sum_tmp = sum_tmp +  conditionMatrix[k2][k] * math.log2(conditionMatrix[k2][k])
                        HYX = HYX + sum_tmp*X_property[k]
                    HYX= -HYX 
                    IGYX = HY - HYX
                    if(HX + HY ==0):
                        SUYX=0
                    else:
                        SUYX = 2*IGYX/(HY + HX)
                    SUY_dict[j] = SUYX
            values = list(SUY_dict.values())
            o = 1
            mean = np.mean(values)
            context_Y = [ key for (key,value) in  SUY_dict.items() if value >= o*mean]
            self.contextMatrix.append(context_Y)
        #Compute distMatrix
        self.distMatrix = [];
        for d in range(D):
            matrix = [] # 2D array for 1 dimension
            for i in range(self.max[d]+1):
                matrix_tmp = []; #1D array for 1 values on the attribute d
                for j in range(self.max[d]+1):
                    dist_sum_all =0;
                    dist_sum = 0;
                    dist_sum2 = 0;
                    for d2 in self.contextMatrix[d]:
                        dist_sum_tmp =0;
                        conditionMatrix = self.conditionProMatrix[(d,d2)]
                        for i_k in range(self.max[d2]):
                            dist_sum_tmp2 =  math.pow(conditionMatrix[i][i_k] -  conditionMatrix[j][i_k],2)
                            dist_sum_tmp = dist_sum_tmp +  dist_sum_tmp2*self.probabilityMatrix[d2][i_k];
                        dist_sum = dist_sum+ dist_sum_tmp
                        dist_sum2 = dist_sum2 + self.max[d2]+1
                    if dist_sum2==0: # toanstt 
                        dist_sum2=1;
                    dist_sum_all = math.sqrt(dist_sum/dist_sum2)
                    matrix_tmp.append(dist_sum_all)
                matrix.append(matrix_tmp)
            self.distMatrix.append(matrix)
    def calculate(self, instance1, instance2):
        distance = 0
        length = len(instance1)
        for x in range(length):
            distance =distance+ math.pow(self.distMatrix[x][instance1[x]][instance2[x]],2)
        return math.sqrt(distance)

    def setUp(self, X, y):
        start = timeit.default_timer()
        self.max = []
        self.X_ = X
        if self.LoaddistMatrixAuto()==False:
            print("Generating disMatrix for DILCA")
            for i in range(len(X[0])):
                self.max.append(max(X[:,i]))
            self.ComputeEtropy(X)
            self.SavedistMatrix()
        return timeit.default_timer() - start




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lshkrepresentatives", 
    version="1.2.3",
    author="nmtoan91",
    author_email="toan_stt@yahoo.com",
    description="LSH-k-Representatives: Mixed categorial and numerical (ordinal and nonordinal) data clustering algorithm algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nmtoan91/lshkrepresentatives",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import math
import numpy as np
from decimal import *


# Calculate the frequency of all values in a data set
def countFreq(dataset):
    length = len(dataset[0, :])
    freq = []
    for i in range(length):
        unique, counts = np.unique(dataset[:, i], return_counts=True)
        temp = np.array(list(zip(unique, counts)))
        freq.append(temp)
    return freq


# calculate the frequency of a value of an attribute k
def freq(Ak, a):
    unique, counts = np.unique(Ak, return_counts=True)
    for i in range(len(unique)):
        if unique[i] == a:
            return Decimal(counts[i].item())
    return 0


# calculate the joint frequency of 2 values of 2 attributes
def condFreq(Ar, Al, xr, xl):
    length = len(Ar)
    count = 0
    for i in range(length):
        if Ar[i] == xr and Al[i] == xl:
            count += 1
    return Decimal(count)


# calculate frequency probability of value a in attribute Ak
def prob(Ak, a):
    unique, counts = np.unique(Ak, return_counts=True)
    n = len(Ak)
    for i in range(len(unique)):
        if unique[i] == a:
            return Decimal(counts[i].item()) / n
    return 0


# calculate conditional probability of value ak of attribute Ak known al of attribute Al: P(ak|al)
def condProb(Ak, Al, ak, al):
    count = 0
    total_al = 0
    for i in range(len(Al)):
        if Al[i] == al:
            total_al += 1
            if Ak[i] == ak:
                count += 1
    if total_al == 0:
        return 0
    else:
        return Decimal(count) / total_al


# calculate joint probability between 2 values ak, al of 2 attributes Ak, Al
def jointProb(Ak, Al, ak, al):
    count = 0
    total = 0
    for i in range(len(Ak)):
        if Ak[i] == ak and Al[i] == al:
            count += 1
        total += 1
    if total == 0:
        return 0
    else:
        return Decimal(count) / total


# generate domain matrix for k attributes of data set
def generateAllDomain(dataset):
    length = len(dataset[0, :])
    A = []
    for i in range(length):
        # A.append(generateDomain(dataset[:, i]))
        A.append(np.unique(dataset[:, i]))
    return A


# generate domain for attribute A
def generateDomain(A):
    dom = []
    for j in A:
        if j not in dom:
            dom.append(j)
    return dom


# calculate mutual information between attribute Ak and Al
def mutualInformation(Ak, Al):
    mutualInfor = 0
    domAk = np.unique(Ak)
    domAl = np.unique(Al)
    for p in range(len(domAk)):
        for q in range(len(domAl)):
            Pkl = jointProb(Ak, Al, domAk[p], domAl[q])
            Pk = prob(Ak, domAk[p])
            Pl = prob(Al, domAl[q])
            if Pkl != 0:
                mutualInfor += Pkl * (Pkl / (Pk * Pl)).log10()
    return mutualInfor


# calculate joint entropy between attribute Ak and Al
def jointEntropy(Ak, Al):
    jointEntro = 0
    domAk = np.unique(Ak)
    domAl = np.unique(Al)
    for p in range(len(domAk)):
        for q in range(len(domAl)):
            Pkl = jointProb(Ak, Al, domAk[p], domAl[q])
            if Pkl != 0:
                jointEntro += Pkl * Pkl.log10()
    return jointEntro * (-1)


# calculate entropy of attribute A based on DILCA
def entropy(A):
    entro = 0
    domA = np.array(np.unique(A))
    for a in domA:
        pa = prob(A, a)
        entro += pa * (pa.log10() / Decimal(2).log10())
    return entro * (-1)


# calculate conditional entropy of attribute Ak given Al based on DILCA
def condEntropy(Ak, Al):
    condEntro = 0
    domAk = np.array(np.unique(Ak))
    domAl = np.array(np.unique(Al))
    for i in range(len(domAl)):
        Pli = prob(Al, domAl[i])
        temp = 0
        for j in range(len(domAk)):
            Pkj_li = condProb(Ak, Al, domAk[j], domAl[i])
            if Pkj_li != 0:
                temp += Pkj_li * (Pkj_li.log10() / Decimal(2).log10())
        condEntro += Pli * temp
    return condEntro * (-1)


# calculate information gain of attribute Ak provided by attribute Al based on DILCA
def inforGain(Ak, Al):
    entro = entropy(Ak)
    conEntro = condEntropy(Ak, Al)
    return entro - conEntro


# calculate Symmetric Uncertainty based on DILCA
def symUncertainty(Ak, Al):
    entro_Ak = entropy(Ak)
    entro_Al = entropy(Al)
    inforG = inforGain(Ak, Al)
    if entro_Al + entro_Ak == 0:
        return 0
    else:
        return (2 * inforG) / (entro_Ak + entro_Al)


# calculate correlation between attributes based on ConDist of Ring
def cor(Ak, Al):
    return inforGain(Ak, Al) / entropy(Ak)


# calculate the interdependence redundancy matrix between attributes of data set based on Jia
def interRedundMatrix(dataset):
    length = len(dataset[0, :])
    # RMatrix = np.array([[0 for x in range(length)] for y in range(length)])
    # RMatrix = np.array([0 for x in range(length)])
    RMatrix = []
    for i in range(length):
        # RMatrix[i, i] = 1
        row = []
        for j in range(length):
            interR_ij = interRedund(dataset[:, i], dataset[:, j])
            row.append(interR_ij)
            # RMatrix[i, j] = interR_ij
            # RMatrix[j, i] = RMatrix[i, j]
        RMatrix.append(row)
    return np.array(RMatrix)


# calculate the average dependence of a data set
def avgDependence(dataset):
    RMatrix = interRedundMatrix(dataset)
    k = len(dataset[0, :])
    depend = 0
    count = 0
    for i in range(0, k - 1):
        for j in range(i + 1, k):
            depend += RMatrix[i][j]
            count += 1
    return Decimal(depend) / count


# calculate the symmetric uncertainty metric between attributes of data set based on DILCA
def symUncertaintyMatrix(dataset):
    length = len(dataset[0, :])
    SMatrix = np.array([[0 for x in range(length)] for y in range(length)])
    for i in range(length):
        SMatrix[i][i] = 1
        for j in range(i + 1, length):
            SMatrix[i][j] = symUncertainty(dataset[:, i], dataset[:, j])
            SMatrix[j][i] = symUncertainty(dataset[:, j], dataset[:, i])
    return SMatrix


# caculate mean of symmetric uncertainty for attribute Ak based on DILCA
def meanSymUncertainty(SMatrix, k):
    mean = 0
    length = len(SMatrix[0, :])
    for i in range(length):
        if i != k:
            mean += SMatrix[k][i]
    return float(mean) / (length - 1)


# calculate the interdependence redundancy between 2 attributes based on Jia
def interRedund(Ak, Al):
    I = mutualInformation(Ak, Al)
    H = jointEntropy(Ak, Al)
    if H == 0:
        R = I
    else:
        R = I / H
    return R

def ProconditionMatrixYX( Y, X,Y_unique, X_unique):
        lenData = len(Y)
        lenX = len(X_unique)
        lenY = len(Y_unique)
        HY = 0
        count_x = 0;
        count_y=0;
        MATRIX = [[0 for i in range(lenX)] for j in  range(lenY)]
        for x in range(lenX):
            for y in range(lenY):
                count_x = count_y =0
                for i in range(lenData):
                    if X[i] == x:
                        count_x = count_x+1
                        if(Y[i] == y):
                            count_y = count_y+1;
                MATRIX[y][x] = count_y/count_x if count_x > 0 else 0
        return MATRIX
